'\"
'\" Copyright (c) 1993 The Regents of the University of California.
'\" Copyright (c) 1994 Sun Microsystems, Inc.
'\"
'\" See the file "license.terms" for information on usage and redistribution
'\" of this file, and for a DISCLAIMER OF ALL WARRANTIES.
'\" 
'\" @(#) unknown.n 1.4 95/05/06 15:19:41
'\" 
.so man.macros
.TH unknown n "" Tcl "Tcl Built-In Commands"
.BS
'\" Note:  do not modify the .SH NAME line immediately below!
.SH NAME
unknown \- Handle attempts to use non-existent commands
.SH SYNOPSIS
\fBunknown \fIcmdName \fR?\fIarg arg ...\fR?
.BE

.SH DESCRIPTION
.PP
This command doesn't actually exist as part of Tcl, but Tcl will
invoke it if it does exist.
If the Tcl interpreter encounters a command name for which there
is not a defined command, then Tcl checks for the existence of
a command named \fBunknown\fR.
If there is no such command, then the interpreter returns an
error.
If the \fBunknown\fR command exists, then it is invoked with
arguments consisting of the fully-substituted name and arguments
for the original non-existent command.
The \fBunknown\fR command typically does things like searching
through library directories for a command procedure with the name
\fIcmdName\fR, or expanding abbreviated command names to full-length,
or automatically executing unknown commands as sub-processes.
In some cases (such as expanding abbreviations) \fBunknown\fR will
change the original command slightly and then (re-)execute it.
The result of the \fBunknown\fR command is used as the result for
the original non-existent command.
.PP
Many extensions redefine the \fBunknown\fR command to include
their own special command processing.  If more than one extension
tries to do this, however, the results conflict.  The standard
\fBunknown\fR command is now set up to handle this correctly.
Each extension can register its own command handler using the
following command:
.CS
unknown_handler \fIname command\fR ?\fIarg arg ...\fR?
.CE
Here, \fIname\fR is a symbolic name identifying the handler.
The \fIcommand\fR string and any extra \fIarg\fR arguments
are integrated into the \fBunknown\fR command.  
.PP
Each time an unknown command is encountered, the list of handlers
is consulted to determine how to handle the command.  If a handler
recognizes the command, then it should load, define, or otherwise
handle the command and return.  Otherwise, it should return
"-code continue", which tells the \fBunknown\fR facility to
continue with another handler.  Handlers registered first are
consulted last.  This way, extensions that are loaded later
have higher priority for handling special cases.
.PP
Usually \fIcommand\fR is the name of a handling procedure.  But
extra \fIarg\fR arguments can be specified when the handler is
registered, and these will be kept as part of the handler.
The actual unknown command and its arguments are appended after
these when the handler is invoked.

.SH KEYWORDS
error, non-existent command
